import { get, writable, type Writable } from "svelte/store"
import {
	API_BASE_ADDR,
	AT_EXP_MS,
	COOKIE_SAME_SITE,
	COOKIE_SECURE,
	CSRF_EXP_MS,
	REFRESH_BUF,
	UUID_REGEX,
	VIEW_COOKIE_DOMAIN,
	VIEW_COOKIE_PATH
} from "./const"
import { goto } from "$app/navigation"
import { usersPagination } from "./pages"

interface User {
	id: string
	username: string
	isAdmin: boolean
	createdAt: Date
	updatedAt: Date
}

interface ApiUserResponse {
	id: string
	username: string
	is_admin: boolean
	created_at: string
	updated_at: string
}

export interface FullNote {
	id: string
	owner: string
	title: string
	content: string
	versionNumber: number
	versionCreatedAt: Date
	isActiveVersion: boolean
	noteCreatedAt: Date
	noteUpdatedAt: Date
}

interface ApiFullNoteResponse {
	note_id: string
	owner_id: string
	title: string
	content: string
	version_number: number
	version_created_at: string
	note_created_at: string
	note_updated_at: string
}

export interface NoteMetadata {
	id: string
	owner: string
	title: string
	updatedAt: Date
}

interface ApiNoteMetadataResponse {
	note_id: string
	owner_id: string
	title: string
	updated_at: string
}

interface NewNoteResponse {
	title: string
	content: string
}

interface VersionMetadata {
	versionID: string
	title: string
	versionNumber: number
	isActive: boolean
	createdAt: Date
}

interface ApiVersionMetadataResponse {
	version_id: string
	title: string
	version_number: number
	created_at: string
}

interface ApiFullVersionResponse {
	version_id: string
	title: string
	content: string
	version_number: number
	created_at: string
}

// Some of these could just be local variables as not all of them are being used globally
export const currentUser: Writable<User | null> = writable(null)
export const currentFullNote: Writable<FullNote | null> = writable(null)
export const availableNotes: Writable<NoteMetadata[] | null> = writable(null)
export const versionHistory: Writable<VersionMetadata[] | null> = writable(null)
export const accessToken: Writable<string | null> = writable(null)
export const csrfToken: Writable<string | null> = writable(null)
export const isPending: Writable<boolean> = writable(false)
export const cError: Writable<string | null> = writable(null)
export const cSuccess: Writable<string | null> = writable(null)

class ApiClient {
	private viewCookieName: string
	private baseUrl: string
	private lastAtUpdate = new Date(0) // Refreshing the page wipes access and CSRF tokens from memory -> Rotation needed
	private lastCsrfUpdate = new Date(0)
	private refreshInProgress = false
	private activeVersion = -1
	private loadedNotesCache = new Map<string, FullNote>()
	private loadedHistoryCache = new Map<string, VersionMetadata[]>()
	private loadedVersionsCache = new Map<string, FullNote>() // Key: noteID + versionID

	constructor(baseUrl: string) {
		this.baseUrl = baseUrl
		this.viewCookieName = "notatest.expires_at"
	}

	private async handleRequest<T>(
		fn: () => Promise<T>,
		options: { useBearerAuth: boolean; suspendGlobalErr: boolean }
	): Promise<T | null> {
		isPending.set(true)
		cError.set(null)
		cSuccess.set(null)

		// NOTE: If `handleResponse` is used, errors thrown from it will be caught here

		try {
			// Only bearers (JWT access tokens) need to be considered here as the refresh token cookies
			// will be automatically included to corresponding requests by browser and CSRF rotation is
			// handled by the `refreshAccessToken` method
			if (options.useBearerAuth) {
				try {
					await this.checkAndRefreshAccessToken()
				} catch (err) {
					console.log("[REQ] Refresh attempt not successful")
					await this.handleLocalLogout()
					throw new Error("Session expired, please authenticate again.")
				}
			}

			return await fn()
		} catch (err) {
			const errMsg = err instanceof Error ? err.message : "Unknown error"

			// The suspension option is handy when we want to display the error inside a modal instead of in a global notification
			if (options.suspendGlobalErr) {
				// Throw the same error to the next handler (should be handled inside the caller component)
				throw new Error(errMsg)
			} else {
				cError.set(errMsg)
			}

			console.log(`[ERR] ${errMsg}`)
		} finally {
			isPending.set(false)
		}

		return null
	}

	// Should be attached to routes that handle authentication with the bearer token (access token)
	private async handleResponse<T>(
		response: Response,
		options: { useBearerAuth: boolean }
	): Promise<T> {
		if (!response.ok) {
			if (response.status === 401 && options.useBearerAuth) {
				// This should never happen due to the token expiration checks we make client-side,
				// but it's still good to have as a fallback
				try {
					console.log("[RES] Unexpected 401 caught, attempting to refresh...")
					await this.checkAndRefreshAccessToken()
				} catch (err) {
					const errMsg = err instanceof Error ? err.message : "Unknown error"
					await this.handleLocalLogout()
					throw new Error(errMsg)
				}
			}

			// Capitalize the error message and display (only if not 401)
			const { error } = await response.json()
			const dError = error[0].toUpperCase() + error.substr(1).toLowerCase() + "."

			throw new Error(dError)
		}

		return await response.json()
	}

	private async checkAndRefreshAccessToken(): Promise<void> {
		// Notably we must check whether we even have an authentication cookie present
		// as that's obviously required for completing the token rotation procedure
		if (this.refreshInProgress) {
			return
		} else if (!this.isAuthCookieValid()) {
			throw new Error("Session expired, please authenticate again.")
		}

		const timeSinceUpdate = Date.now() - this.lastAtUpdate.getTime()
		const needsRefresh = timeSinceUpdate > AT_EXP_MS - REFRESH_BUF

		if (needsRefresh) {
			console.log(`[AUTH] Running token refresh attempt (timeSinceUpdate=${timeSinceUpdate})`)

			this.refreshInProgress = true
			await this.refreshAccessToken()
			this.refreshInProgress = false
		}
	}

	private async refreshAccessToken(): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/cookie/refresh`, {
					method: "POST",
					headers: {
						...(await this.getCsrfHeader())
					},
					credentials: "include"
				})

				if (!response.ok) {
					await this.handleLocalLogout()
					throw new Error("Refreshing token failed.")
				}

				const { access_token: newToken } = await response.json()
				accessToken.set(newToken)
				this.lastAtUpdate = new Date()
			},
			{ useBearerAuth: false, suspendGlobalErr: false }
		)
	}

	private async getCsrfHeader(): Promise<HeadersInit> {
		let token = get(csrfToken)
		const timeSinceUpdate = Date.now() - this.lastCsrfUpdate.getTime()
		const needsRefresh = timeSinceUpdate > CSRF_EXP_MS - REFRESH_BUF

		if (!token || needsRefresh) {
			console.log("[AUTH] Refreshing CSRF token")
			await this.refreshCsrfToken()
			token = get(csrfToken)
		}

		return { "X-Csrf-Token": token || "" }
	}

	private async refreshCsrfToken(): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/cookie/csrf`, {
					credentials: "include"
				})

				if (!response.ok) {
					await this.handleLocalLogout()
					throw new Error("Fetching CSRF token failed.")
				}

				const newToken = response.headers.get("X-Csrf-Token")
				csrfToken.set(newToken)
				this.lastCsrfUpdate = new Date()
			},
			{ useBearerAuth: false, suspendGlobalErr: false }
		)
	}

	private async handleLocalLogout(): Promise<void | null> {
		this.lastAtUpdate = new Date(0)
		this.lastCsrfUpdate = new Date(0)

		currentUser.set(null)
		accessToken.set(null)
		csrfToken.set(null)

		this.deleteViewCookie()

		await goto("/login")
	}

	private getAuthHeader(): HeadersInit {
		const token = get(accessToken)
		return { Authorization: `Bearer ${token}` }
	}

	private isAuthCookieValid(): boolean {
		const now = new Date()
		const viewCookie = this.getCookieValue(this.viewCookieName)

		if (!viewCookie) {
			console.log("[AUTH] View cookie not found")
			return false
		}

		const value = viewCookie.split("=")[1]
		const expirationTimestamp = parseInt(value, 10)

		if (isNaN(expirationTimestamp)) {
			console.log(`[ERR] Invalid view cookie expiration timestamp: ${value}`)
			return false
		}

		return now < new Date(expirationTimestamp * 1000)
	}

	private deleteViewCookie() {
		const viewCookie = this.getCookieValue(this.viewCookieName)
		if (!viewCookie) {
			return
		}

		// Overwrite the view cookie with details that match with the real cookie
		document.cookie = `${this.viewCookieName}=;path=${VIEW_COOKIE_PATH};domain=${VIEW_COOKIE_DOMAIN};expires=Thu, 01 Jan 1970 00:00:00 GMT;${COOKIE_SECURE ? "secure;" : ""}sameSite=${COOKIE_SAME_SITE}`
	}

	private getCookieValue(cookieName: string): string | null {
		const cookieString = document.cookie
		const cookies = cookieString.split(";").map((cookie) => cookie.trim())
		const cookieValue = cookies.find((cookie) => cookie.startsWith(`${cookieName}=`))

		if (!cookieValue) {
			return null
		}

		return cookieValue
	}

	private deserializeUser(apiResponse: ApiUserResponse): User {
		return {
			id: apiResponse.id,
			username: apiResponse.username,
			isAdmin: apiResponse.is_admin,
			createdAt: new Date(apiResponse.created_at),
			updatedAt: new Date(apiResponse.updated_at)
		}
	}

	private deserializeNoteMetadatas(apiResponses: ApiNoteMetadataResponse[]): NoteMetadata[] {
		return apiResponses.map((res) => {
			return {
				id: res.note_id,
				owner: res.owner_id,
				title: res.title,
				updatedAt: new Date(res.updated_at)
			}
		})
	}

	private deserializeFullNote(apiResponse: ApiFullNoteResponse): FullNote {
		return {
			id: apiResponse.note_id,
			owner: apiResponse.owner_id,
			title: apiResponse.title,
			content: apiResponse.content,
			versionNumber: apiResponse.version_number,
			versionCreatedAt: new Date(apiResponse.version_created_at),
			isActiveVersion: true, // This endpoint serves only the latest version
			noteCreatedAt: new Date(apiResponse.note_created_at),
			noteUpdatedAt: new Date(apiResponse.note_updated_at)
		}
	}

	private deserializeVersionMetadatas(
		apiResponses: ApiVersionMetadataResponse[]
	): VersionMetadata[] {
		return apiResponses.map((res) => {
			return {
				versionID: res.version_id,
				title: res.title,
				versionNumber: res.version_number,
				isActive: this.activeVersion === res.version_number,
				createdAt: new Date(res.created_at)
			}
		})
	}

	private joinDeserializedVersion(
		noteID: string,
		apiResponse: ApiFullVersionResponse
	): FullNote | null {
		// Cache lookups are safe here due to this always being called *after* fetching the actual `FullNote`
		const cachedNote = this.loadedNotesCache.get(noteID)

		if (!cachedNote) {
			return null
		}

		return {
			id: cachedNote.id,
			owner: cachedNote.owner,
			title: apiResponse.title,
			content: apiResponse.content,
			versionNumber: apiResponse.version_number,
			versionCreatedAt: new Date(apiResponse.created_at),
			isActiveVersion: cachedNote.versionNumber === apiResponse.version_number,
			noteCreatedAt: cachedNote.noteCreatedAt,
			noteUpdatedAt: cachedNote.noteUpdatedAt
		}
	}

	public async register(username: string, password: string): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/signup`, {
					method: "POST",
					headers: { "Content-Type": "application/json" },
					body: JSON.stringify({ username, password })
				})

				// Can't overwrite the function parameter
				const data = await this.handleResponse<{
					id: string
					username: string
				}>(response, { useBearerAuth: false })

				console.log(`[USER] Registration of user '${data.username}' successful`)

				await goto("/login")
			},
			{ useBearerAuth: false, suspendGlobalErr: false }
		)
	}

	public async login(username: string, password: string): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/login`, {
					method: "POST",
					headers: { "Content-Type": "application/json" },
					body: JSON.stringify({ username, password }),
					credentials: "include"
				})

				const { access_token: token } = await this.handleResponse<{
					access_token: string
				}>(response, { useBearerAuth: false })

				accessToken.set(token)
				this.lastAtUpdate = new Date()

				goto("/")
			},
			{ useBearerAuth: false, suspendGlobalErr: false }
		)
	}

	public async logout(): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/logout`, {
					method: "POST",
					headers: {
						...this.getAuthHeader()
					},
					credentials: "include"
				})

				if (response.status === 204) {
					console.log("[USER] Logout successful")
					await this.handleLocalLogout()
					return
				}

				await this.handleResponse<void>(response, { useBearerAuth: true })
				await this.handleLocalLogout()
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async getCurrentUser(): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/me`, {
					headers: {
						...this.getAuthHeader()
					}
				})
				const data = await this.handleResponse<ApiUserResponse>(response, { useBearerAuth: true })
				const user = this.deserializeUser(data)

				currentUser.set(user)
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async updateCurrentUserPassword(
		oldPassword: string,
		newPassword: string
	): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const data = {
					old_password: oldPassword,
					new_password: newPassword
				}
				const response = await fetch(`${this.baseUrl}/auth/owner`, {
					method: "PUT",
					headers: {
						...this.getAuthHeader(),
						"Content-Type": "application/json"
					},
					credentials: "include",
					body: JSON.stringify(data)
				})

				const { access_token: token, user } = await this.handleResponse<{
					access_token: string
					user: User
				}>(response, { useBearerAuth: true })

				accessToken.set(token)
				currentUser.set(user || null)
				this.lastAtUpdate = new Date()
			},
			{ useBearerAuth: true, suspendGlobalErr: true } // Error displayed inside the settings modal
		)
	}

	public async deleteCurrentUser(password: string): Promise<void | null> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/owner`, {
					method: "DELETE",
					headers: {
						...this.getAuthHeader(),
						"Content-Type": "application/json"
					},
					body: JSON.stringify({ password })
				})

				if (response.status === 204) {
					console.log("[USER] Deletion successful")
					await this.handleLocalLogout()
					return
				}

				await this.handleResponse<void>(response, { useBearerAuth: true })
				await this.handleLocalLogout()
			},
			{ useBearerAuth: true, suspendGlobalErr: true } // Error displayed inside the settings modal
		)
	}

	public async adminListAll(): Promise<User[] | null> {
		const user = get(currentUser)
		if (!user || !user.isAdmin) {
			throw new Error("Admin privileges required.")
		}

		return this.handleRequest(
			async () => {
				const params = new URLSearchParams()
				params.append("limit", `${get(usersPagination).pageSize}`)
				params.append("offset", `${get(usersPagination).currentPage}`)
				const response = await fetch(`${this.baseUrl}/auth/admin/all?${params}`, {
					headers: {
						...this.getAuthHeader()
					}
				})

				const users = await this.handleResponse<User[]>(response, { useBearerAuth: true })
				console.log(`[ADMIN] Got ${users.length} user results`)

				return users
			},
			{ useBearerAuth: true, suspendGlobalErr: true } // Error displayed inside the settings modal
		)
	}

	public async adminDeleteUser(userID: string): Promise<void | null> {
		const user = get(currentUser)
		if (!user || !user.isAdmin) {
			throw new Error("Admin privileges required.")
		}

		if (!UUID_REGEX.test(userID)) {
			throw new Error("Invalid user ID format.")
		}

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/admin/${userID}`, {
					method: "DELETE",
					headers: {
						...this.getAuthHeader()
					}
				})

				if (response.status === 204) {
					console.log(`[ADMIN] Deletion of user '${userID}' successful`)
					return
				}

				await this.handleResponse<void>(response, { useBearerAuth: true })
			},
			{ useBearerAuth: true, suspendGlobalErr: true } // Error displayed inside the settings modal
		)
	}

	public async listNotes(): Promise<NoteMetadata[] | null> {
		return this.handleRequest(
			async () => {
				const params = new URLSearchParams()
				params.append("limit", `${get(usersPagination).pageSize}`)
				params.append("offset", `${get(usersPagination).currentPage}`)
				const response = await fetch(`${this.baseUrl}/notes?${params}`, {
					headers: {
						...this.getAuthHeader()
					}
				})

				let notes: NoteMetadata[] = []
				let data = await this.handleResponse<ApiNoteMetadataResponse[]>(response, {
					useBearerAuth: true
				})

				if (data) {
					notes = this.deserializeNoteMetadatas(data)
				}

				console.log(`[NOTE] Got ${notes.length} note metadata results`)

				return notes
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async createNote(): Promise<NewNoteResponse | null> {
		// NOTE: The initial note version doesn't allow any user input, the first user-made modification
		//       is applied through the version creation endpoint

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/notes`, {
					method: "POST",
					headers: {
						...this.getAuthHeader()
					}
				})
				return await this.handleResponse<NewNoteResponse>(response, { useBearerAuth: true })
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async getActiveFullNote(noteID: string, fetchRemote: boolean): Promise<FullNote | null> {
		if (!UUID_REGEX.test(noteID)) {
			throw new Error("Invalid note ID format.")
		}

		// Attempt cache lookup only if we didn't just push new updates
		if (!fetchRemote) {
			const cachedNote = this.loadedNotesCache.get(noteID)
			if (cachedNote != null) {
				this.activeVersion = cachedNote.versionNumber
				return cachedNote
			}
		}

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/notes/${noteID}`, {
					headers: {
						...this.getAuthHeader()
					}
				})

				const data = await this.handleResponse<ApiFullNoteResponse>(response, {
					useBearerAuth: true
				})
				const note = this.deserializeFullNote(data)

				console.log(`[CACHE] Storing ${noteID}`)
				this.loadedNotesCache.set(noteID, note)
				this.activeVersion = note.versionNumber

				return note
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async deleteNote(noteID: string): Promise<void | null> {
		if (!UUID_REGEX.test(noteID)) {
			throw new Error("Invalid note ID format.")
		}

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/notes/${noteID}`, {
					method: "DELETE",
					headers: {
						...this.getAuthHeader()
					}
				})

				if (response.status === 204) {
					console.log("[NOTE] Deletion successful")
					return
				}

				await this.handleResponse<void>(response, { useBearerAuth: true })
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async getNoteHistory(
		noteID: string,
		fetchRemote: boolean
	): Promise<VersionMetadata[] | null> {
		if (!UUID_REGEX.test(noteID)) {
			throw new Error("Invalid note ID format.")
		}

		if (!fetchRemote) {
			const cachedVersions = this.loadedHistoryCache.get(noteID)
			if (cachedVersions != null) {
				return cachedVersions
			}
		}

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/notes/${noteID}/versions`, {
					headers: {
						...this.getAuthHeader()
					}
				})

				const data = await this.handleResponse<ApiVersionMetadataResponse[]>(response, {
					useBearerAuth: true
				})
				const versions = this.deserializeVersionMetadatas(data)

				this.loadedHistoryCache.set(noteID, versions)
				console.log(`[VER] Got and cached ${versions.length} version metadata results`)

				return versions
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async createVersion(noteID: string, title: string, content: string): Promise<void | null> {
		if (!UUID_REGEX.test(noteID)) {
			throw new Error("Invalid note ID format.")
		}

		// NOTE: Title's length limit is applied in the UI component, so we don't need to worry about it here

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/notes/${noteID}/versions`, {
					method: "POST",
					headers: {
						...this.getAuthHeader(),
						"Content-Type": "application/json"
					},
					body: JSON.stringify({ title, content })
				})

				if (response.status === 204) {
					console.log("[VER] Creation successful")
					return
				}

				await this.handleResponse<void>(response, { useBearerAuth: true })
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}

	public async getFullVersion(noteID: string, versionID: string): Promise<FullNote | null> {
		if (!UUID_REGEX.test(noteID)) {
			throw new Error("Invalid note ID format.")
		}

		if (!UUID_REGEX.test(versionID)) {
			throw new Error("Invalid version ID format.")
		}

		// NOTE: No need to explicitly prevent attempting a cache hit as versions aren't editable

		// TODO: If accessing the active version, don't attempt to hit the cache,
		//       but instead load the contents from the currently active full note

		const cachedVersion = this.loadedVersionsCache.get(noteID + versionID)
		if (cachedVersion != null) {
			return cachedVersion
		}

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/notes/${noteID}/${versionID}`, {
					headers: {
						...this.getAuthHeader()
					}
				})

				const data = await this.handleResponse<ApiFullVersionResponse>(response, {
					useBearerAuth: true
				})
				const version = this.joinDeserializedVersion(noteID, data)

				if (!version) {
					return null
				}

				console.log(`[CACHE] Storing [${noteID}, ${versionID}]`)
				this.loadedVersionsCache.set(noteID + versionID, version)

				return version
			},
			{ useBearerAuth: true, suspendGlobalErr: false }
		)
	}
}

export const apiClient = new ApiClient(API_BASE_ADDR)
